/* Include files */
#include <windows.h>
#include <commctrl.h>
#include <stdlib.h>
#include <stdio.h>
#include <setupapi.h>					
#include <dbt.h>
#define _MT
#include <process.h>
extern "C" {
#include <hidsdi.h>	/* note: part of Windows DDK */
#include <hidpi.h>	/* note: also part of Windows DDK */
}
#include "usbtherm.h"
#include "resource.h"
#include "time.h"


/* global variables */
HINSTANCE		hGInstance	= NULL;		/* handle to application instance */
char	*build_time		= __TIME__;		/* build time string */
char	*build_date		= __DATE__;		/* build date string */
BOOL endthread = TRUE;					/* thread end message */
BOOL TMThreadActive = FALSE;

char sFormat[1000];

HICON hRedIcon, hBlueIcon, hGrayIcon;

/* begin software listing */

//*********************************************************
//
//	Function: WinMain
//	Purpose: Initial starting function for windows apps
//	Parameters: hInstance - handle to an instance of this application
//				hPrevInstance - unused for Win32 apps
//				lpCmdLine - string containing command line parameters
//				nShowCmd - Command of how to show initial window
//
//*********************************************************

int PASCAL WinMain (HINSTANCE hInstance,
					HINSTANCE hPrevInstance,
					LPSTR lpCmdLine,
					int nShowCmd)
{

INITCOMMONCONTROLSEX	ccex;
	
	if (strlen(lpCmdLine))
	{
		strcpy(sFormat, lpCmdLine);
		strcat(sFormat, "\n");
	}
	else
		strcpy(sFormat, "Temperature: %dC (%dF).\n");
	

	hGInstance = hInstance;			// store global copy of instance handle

	ccex.dwSize = sizeof(ccex);
	ccex.dwICC = ICC_PROGRESS_CLASS;


	/* Initialize the common controls used in this app */
	if(InitCommonControlsEx(&ccex) == FALSE)
	{
		MessageBox(NULL, "Failed to Initialize Application", "Failure!", MB_ICONSTOP);
		return EXIT_FAILURE;
	}


	if(DialogBox(hInstance, MAKEINTRESOURCE(IDD_MAIN), NULL, (DLGPROC) DlgProc) == -1)
	{
		MessageBox(NULL, "Failed to create main dialog box.","Failure!", MB_ICONSTOP);
		return EXIT_FAILURE;
	}

	return	EXIT_SUCCESS;
	}


//*********************************************************
//
//	Function: bMainDlgProc
//	Purpose: This function gets called to handle the main
//			dialog box
//	Parameters: hDlg - handle to this dialog box
//				message - windows messages 
//				wParam - windows message parameters
//				lParam - windows message parameters
//	Returns:
//
//*********************************************************

static HANDLE HidDevHandle=0;						/* Handle to HID device */
static BOOLEAN HidAttached;
static HDEVNOTIFY *hDevNotify;					/* notification results */

static NOTIFYICONDATA nid;			// struct for Shell_NotifyIcon args

static old_tk = -1;

BOOL CALLBACK DlgProc(HWND hwndDlg, UINT uMsg, WPARAM wParam, LPARAM lParam)
{



	/* parse windows messages sent to the dialog box */
	switch(uMsg)	{
	
		/* dialog box initialization */
		case WM_INITDIALOG:
			{

				// Load icons

				hRedIcon = LoadIcon(hGInstance, MAKEINTRESOURCE(IDI_ICON1));
				hBlueIcon = LoadIcon(hGInstance, MAKEINTRESOURCE(IDI_ICON2));
				hGrayIcon = LoadIcon(hGInstance, MAKEINTRESOURCE(IDI_ICON3));

				// Add tray icon

				memset(&nid, 0 , sizeof(nid));
				nid.cbSize = sizeof(nid);
				nid.uID = 0;
				nid.hWnd = hwndDlg;
				strcpy(nid.szTip, "***");
				nid.uCallbackMessage = WM_USER+0;
				nid.uFlags = NIF_ICON|NIF_TIP|NIF_MESSAGE;
				nid.hIcon = hGrayIcon;
				Shell_NotifyIcon(NIM_ADD, &nid);

				SendMessage(hwndDlg, WM_SETICON, 1, (LPARAM)nid.hIcon);

				/* Check to see if the HID Thermometer is attached to the system */
				HidAttached = bOpenHidDevice(&HidDevHandle, VID, PID);

				if(HidAttached == TRUE)
				{
					CloseHandle(HidDevHandle);
					endthread = FALSE;
					_beginthread(TMeasure, 0, hwndDlg);
				}

				/* Set up notifications of HID device attach and detach events */
				if( bHidDeviceNotify(hwndDlg, hDevNotify) == FALSE)
					MessageBox(hwndDlg, "Failed to Register Device Notifications", "ERROR", MB_ICONSTOP);				

				
				SetTimer(hwndDlg, 1, 3000, 0);      

				break;
			}
			
		/* these are the various controls in the dialog window */
		case WM_COMMAND:
		{
			break;
		}

		case WM_TIMER:
		{
			KillTimer(hwndDlg, 1);      
			ShowWindow(hwndDlg, SW_HIDE);
			break; 
		}

		case WM_SYSCOMMAND:
		{
			if (wParam == 61472)
			{
				SetTimer(hwndDlg, 1, 1, 0); 
//				return TRUE;
			}
//			else
//				return TRUE; 
			break;
		}

		case WM_USER+0: // Tray notification
		{
			KillTimer(hwndDlg, 1);      

			if (lParam == WM_RBUTTONDOWN)
			{
				POINT pt;
				GetCursorPos(&pt);
				HMENU hMenu = LoadMenu(hGInstance, MAKEINTRESOURCE(IDR_MENU1));
				hMenu = GetSubMenu(hMenu, 0);
				SetForegroundWindow(hwndDlg);
				int iChoice = TrackPopupMenu(hMenu, TPM_RETURNCMD|TPM_NONOTIFY, pt.x, pt.y, 0,nid.hWnd, NULL);
				PostMessage(hwndDlg, WM_NULL, 0, 0);

				if (iChoice==1)
				{
					ShowWindow(hwndDlg, IsWindowVisible(hwndDlg) ? SW_HIDE:SW_RESTORE);
					SetForegroundWindow(hwndDlg);
//					SetForegroundWindow(hwndDlg);	
				}
				else
				if (iChoice==2)
				{
					SendMessage(hwndDlg, WM_CLOSE,0,0);
				}

			}
			else
			if (lParam == WM_LBUTTONDOWN)
			{
				ShowWindow(hwndDlg, IsWindowVisible(hwndDlg) ? SW_HIDE:SW_RESTORE);
				SetForegroundWindow(hwndDlg);
//				SetForegroundWindow(hwndDlg);	
			}				

			break;
		}

		case WM_USER+1: // Temperature, K
		{
			int tk = (int)wParam;
			int tc = tk-273;
			int tf = tc*1.8 +32;

			if (tk != old_tk)
			{
				// Display
				char s[10] = "***";
				if (tk>0)
					sprintf(s,"%d C", tc );
//					sprintf(s,"%dC, %dF", tc, tf );
				SetDlgItemText(hwndDlg, IDC_TEMPERATURE, s);
	//			SetWindowText(hwndDlg, s);

				// Tip
				strcpy(nid.szTip, s);
				if (tk==0)
					nid.hIcon = hGrayIcon;
				else
				if (tc>0)
					nid.hIcon = hRedIcon;
				else
				if (tc<=0)
					nid.hIcon = hBlueIcon;

				Shell_NotifyIcon(NIM_MODIFY, &nid);

				FILE* f;
			
				// Text

				f = fopen("temperature.txt", "w");

				if (f)
				{
					if (tk>0)
					{
						fprintf(f, sFormat, tc, tf );
					}
					fclose (f);
				}

				f = fopen("celsius.txt", "w");

				if (f)
				{
					if (tk>0)
						fprintf(f,"%d", tc );
					fclose (f);
				}

				f = fopen("fahrenheit.txt", "w");

				if (f)
				{
					if (tk>0)
						fprintf(f,"%d", tf );
					fclose (f);
				}

				// APRS

				f = fopen("ui-view.txt", "w");

				if (f)
				{

					time_t t1;
					time(&t1);
					tm* t2 = localtime(&t1);
					const char* months[] = {"Jan","Feb","Mar","Apr","May","Jun","Jul","Aug","Sep","Oct","Nov","Dec"};
					
					fprintf(f,"%s %02d %04d %02d:%02d\n", 
						months[t2->tm_mon], t2->tm_mday, t2->tm_year+1900, t2->tm_hour, t2->tm_min );

					if (tk>0)
						fprintf(f,"000/000t%03d\n", tf );
					else
						fprintf(f,"000/000t...\n");

					fclose (f);
				}


			}

			old_tk = tk;

			break;
		}

		/* When a (meaning ANY) HID device is attached or detached
		the application will receive a WM_DEVICECHANGE message.  The
		application then needs to try to open a handle to the device.
		If it succeeds, then the device is attached.  If it fails, then
		the device was detached.*/
		case WM_DEVICECHANGE:
		{
			/* try to open a handle to the device */
			HidAttached = bOpenHidDevice(&HidDevHandle, VID, PID);
			CloseHandle(HidDevHandle);			

			
			/* take the appropriate action for the message */
			switch(LOWORD(wParam)) {
						
				/* HID device arrival */
				case DBT_DEVICEARRIVAL:
				{
					/* if attached then start temperature measurements */					
					if(endthread == TRUE && HidAttached == TRUE)
					{
						endthread = FALSE;
						_beginthread(TMeasure, 0, hwndDlg);
					}
					break;
				}

				/* HID device removal */
				case DBT_DEVICEREMOVECOMPLETE:
				{
					/* if failed, then stop temperature measurements */
					if(endthread == FALSE && HidAttached == FALSE)
					{
						/* end the IN endpoint thread */
						endthread = TRUE;
						/* reset the temperature text readout */
						SendMessage(hwndDlg, WM_USER+1, 0,0);
					}

					break;
				}

				
				/* HID Device Change - Note that Win2k tends to use this messages
				instead of the previous two. */
				case DBT_DEVNODES_CHANGED:
				{
				
					/* if success then start temperature measurements */					
					if(endthread == TRUE && HidAttached == TRUE)
					{
						endthread = FALSE;
						_beginthread(TMeasure, 0, hwndDlg);
					}
					
					/* if failed, then stop temperature measurements */
					else if(endthread == FALSE && HidAttached == FALSE)
					{
						/* end the IN endpoint thread */
						endthread = TRUE;
						/* reset the temperature text readout */
						SendMessage(hwndDlg, WM_USER+1, 0,0);
					}					
					
					break;
				}



				default:
					break;
			}
			
			break;
		}


		/* little "X" in upper right of window */
		case WM_CLOSE:
			{

				// Add tray icon

				memset(&nid, 0 , sizeof(nid));
				nid.cbSize = sizeof(nid);
				nid.uID = 0;
				nid.hWnd = hwndDlg;
				Shell_NotifyIcon(NIM_DELETE, &nid);
			
				// Clean files

				FILE* f;
			
				f = fopen("temperature.txt", "w");
				if (f)
					fclose (f);
				f = fopen("celsius.txt", "w");
				if (f)
					fclose (f);
				f = fopen("fahrenheit.txt", "w");
				if (f)
					fclose (f);
				f = fopen("ui-view.txt", "w");
				if (f)
					fclose (f);


				/* stop the temperature measurement thread */
				endthread = TRUE;

				/* wait for temperature measurement thread to stop */
				if(TMThreadActive == TRUE)
				{
					PostMessage(hwndDlg, WM_CLOSE, 0, 0);
					break;
				}
				/* end HID device plug/unplug notifications */
				UnregisterDeviceNotification( hDevNotify );

				/* end the dialog box */
				EndDialog(hwndDlg, 0);
				break;
			}
	
		default:
			break;
		}

	
	return	FALSE;
	}


/**********************************************************
*
*	Function TMeasure
*	Purpose: This thread takes temperature measurements from
*		the IN endpoint and reports them to the main window
*	Parameters: none
*	Returns: none
*
**********************************************************/

void TMeasure(LPVOID lpParameter)
{

	HWND			hDlg = (HWND) lpParameter;
	HANDLE			hDevice;
	unsigned long	numBytesReturned;
	unsigned char	inbuffer[9];		/* input buffer*/
	BOOL			bResult;
	HIDP_CAPS		Capabilities;
	PHIDP_PREPARSED_DATA		HidParsedData;
	OVERLAPPED		HidOverlapped;
	HANDLE			ReportEvent;
	short			temperature = 0;	/* temperature measurement */


	TMThreadActive = TRUE;

	
	

	/* Open the HID device handle */
	if(bOpenHidDevice( &hDevice, VID, PID) == TRUE)
	{
		/* get a handle to a buffer that describes the device's capabilities.  This
		line plus the following two lines of code extract the report length the
		device is claiming to support */
		HidD_GetPreparsedData(hDevice, &HidParsedData);
		
		/* extract the capabilities info */
		HidP_GetCaps( HidParsedData ,&Capabilities);
		
		/* Free the memory allocated when getting the preparsed data */
		HidD_FreePreparsedData(HidParsedData);		
		
		/* Create a new event for report capture */
		ReportEvent = CreateEvent(NULL, TRUE, FALSE, NULL);

		/* fill the HidOverlapped structure so that Windows knows which
		event to cause when the device sends an IN report */
		HidOverlapped.hEvent = ReportEvent;
		HidOverlapped.Offset = 0;
		HidOverlapped.OffsetHigh = 0;

/*
		/* Use WriteFile to send an output report to the HID device.  In this
		case we are turning on the READY LED on the target * /
		outbuffer[0] = 0;	/* this is used as the report ID * /
		outbuffer[1] = 1;	/* this flag turns on the LED * /

		bResult = WriteFile(	hDevice, 
								&outbuffer[0], 
								Capabilities.OutputReportByteLength, 
								&numBytesReturned, 
								(LPOVERLAPPED) &HidOverlapped);
		

		bResult = WaitForSingleObject(ReportEvent, 200);
					
*/
		while(endthread == FALSE)
		{

			/* get a temperature report from the HID device.  Note that
			this call to ReadFile only kicks off the reading of the report.
			If the input buffer is full then this routine will cause an
			event to happen and data will be presented immediately.  If not
			the event will occur when the HID decide provides a HID report
			on the IN endpoint. */		
			bResult = ReadFile(	hDevice,							/* handle to device */
								&inbuffer[0],						/* IN report buffer to fill */
								Capabilities.InputReportByteLength,	/* input buffer size */ 
								&numBytesReturned,					/* returned buffer size */ 
								(LPOVERLAPPED) &HidOverlapped );	/* long pointer to an OVERLAPPED structure */

			/* wait for IN report event.  Note that if a report does not occur
			in the time set by the 'dwMilliseconds' parameter, then this function
			returns with a FALSE result and the HID device did not provide a 
			report.  If the function returns TRUE, then a report did occur and the
			data is ready to be read from the buffer specified in the ReadFile
			function call.*/

				bResult = WaitForSingleObject( 	ReportEvent, 
												300);	
				/* if the transaction timed out, then we have to manually cancel
				the request */
				if(bResult == WAIT_TIMEOUT || bResult == WAIT_ABANDONED)
					CancelIo(&hDevice);

				/* Process in the input data.  Note that the first byte (i.e.
				inbuffer[0] is the report ID for the HID report.  We can just
				ignore this value.  The first data byte of interest is
				inbuffer[1] */

				if(bResult == WAIT_OBJECT_0)
				{										
				
					/* check for valid temperature and display it */
					temperature = (short)inbuffer[1] + ((short)inbuffer[2] << 8);
					if(temperature < 1000 && temperature > 0)
					{

						SendMessage(hDlg, WM_USER+1, temperature, 0);
					}
					else
					{

						SendMessage(hDlg, WM_USER+1, 0,0);
					}
				
				}

		
				Sleep(2000);
		}
/*
		/* Use WriteFile to send an output report to the HID device.  In this
		case we are turning off the READY LED on the target * /
		outbuffer[0] = 0;	/* this is used as the report ID * /
		outbuffer[1] = 0;	/* this flag turns on the LED * /

		bResult = WriteFile(	hDevice, 
								&outbuffer[0], 
								Capabilities.OutputReportByteLength, 
								&numBytesReturned, 
								(LPOVERLAPPED) &HidOverlapped);
		
		bResult = WaitForSingleObject(ReportEvent, 200);

*/
		/* close the HID device handle */
		CloseHandle(hDevice);
	}

	TMThreadActive = FALSE;

}


/**********************************************************
*
*	Function: bOpenHidDevice
*	Purpose: tries to open a HID device based on VID and PID
*	Parameters: vid - HID device's vendor ID
*				pid - HID device's product ID
*				HidDevHandle - pointer to a handle to the HID device
*	Returns: TRUE, if device is found
*			 FALSE, if device is not found
*
**********************************************************/

BOOL bOpenHidDevice(HANDLE *HidDevHandle, USHORT vid, USHORT pid)
{
static GUID HidGuid;						/* HID Globally Unique ID: windows supplies us with this value */
HDEVINFO HidDevInfo;						/* handle to structure containing all attached HID Device information */
SP_DEVICE_INTERFACE_DATA devInfoData;		/* Information structure for HID devices */
BOOLEAN Result;								/* result of getting next device information structure */
DWORD Index;								/* index of HidDevInfo array entry */
DWORD DataSize;								/* size of the DeviceInterfaceDetail structure */		
BOOLEAN GotRequiredSize;					/* 1-shot got device info data structure size flag */
PSP_DEVICE_INTERFACE_DETAIL_DATA detailData;/* device info data */
DWORD RequiredSize;							/* size of device info data structure */
BOOLEAN DIDResult;							/* get device info data result */
HIDD_ATTRIBUTES HIDAttrib;					/* HID device attributes */

				
				/* initialize variables */
				GotRequiredSize = FALSE;



				/* 1) Get the HID Globally Unique ID from the OS */
				HidD_GetHidGuid(&HidGuid);


				/* 2) Get an array of structures containing information about
				all attached and enumerated HIDs */
				HidDevInfo = SetupDiGetClassDevs(	&HidGuid, 
													NULL, 
													NULL, 
													DIGCF_PRESENT | DIGCF_INTERFACEDEVICE);


				/* 3) Step through the attached device list 1 by 1 and examine
				each of the attached devices.  When there are no more entries in
				the array of structures, the function will return FALSE. */
				
				Index = 0;									/* init to first index of array */
				devInfoData.cbSize = sizeof(devInfoData);	/* set to the size of the structure
															that will contain the device info data */
				
				do {
					/* Get information about the HID device with the 'Index' array entry */
					Result = SetupDiEnumDeviceInterfaces(	HidDevInfo, 
															0, 
															&HidGuid, 
															Index, 
															&devInfoData);
					
					/* If we run into this condition, then there are no more entries
					to examine, we might as well return FALSE at point */
					if(Result == FALSE)
					{
						/* free the memory allocated for DetailData */
						if(detailData != NULL)
							free(detailData);
						
						/* free HID device info list resources */
						SetupDiDestroyDeviceInfoList(HidDevInfo);
						
						return FALSE;
					}


					if(GotRequiredSize == FALSE)
					{
						/* 3) Get the size of the DEVICE_INTERFACE_DETAIL_DATA
						structure.  The first call will return an error condition, 
						but we'll get the size of the strucure */
						DIDResult = SetupDiGetDeviceInterfaceDetail(	HidDevInfo,
																	&devInfoData,
																	NULL,
																	0,
																	&DataSize,
																	NULL);
						GotRequiredSize = TRUE;

						/* allocate memory for the HidDevInfo structure */
						detailData = (PSP_DEVICE_INTERFACE_DETAIL_DATA) malloc(DataSize);
						
						/* set the size parameter of the structure */
						detailData->cbSize = sizeof(SP_DEVICE_INTERFACE_DETAIL_DATA);
					}
							
						
					/* 4) Now call the function with the correct size parameter.  This 
					function will return data from one of the array members that 
					Step #2 pointed to.  This way we can start to identify the
					attributes of particular HID devices.  */
					DIDResult = SetupDiGetDeviceInterfaceDetail(	HidDevInfo,
																&devInfoData,
																detailData,
																DataSize,
																&RequiredSize,
																NULL);
						

					/* 5) Open a file handle to the device.  Make sure the
					attibutes specify overlapped transactions or the IN
					transaction may block the input thread. */
					*HidDevHandle = CreateFile( detailData->DevicePath,
												GENERIC_READ | GENERIC_WRITE,
												FILE_SHARE_READ | FILE_SHARE_WRITE,
												NULL,
												OPEN_EXISTING,
												FILE_FLAG_OVERLAPPED,
												NULL);

						
					/* 6) Get the Device VID & PID to see if it's the device we want */
					if(*HidDevHandle != INVALID_HANDLE_VALUE)
					{
						HIDAttrib.Size = sizeof(HIDAttrib);
						HidD_GetAttributes(	*HidDevHandle, &HIDAttrib);

						if((HIDAttrib.VendorID == VID) && (HIDAttrib.ProductID == PID))
						{		
							/* free the memory allocated for DetailData */
							if(detailData != NULL)
								free(detailData);
							
							/* free HID device info list resources */
							SetupDiDestroyDeviceInfoList(HidDevInfo);

							return TRUE;	/* found HID device */
						}
						
						/* 7) Close the Device Handle because we didn't find the device
						with the correct VID and PID */
						CloseHandle(*HidDevHandle);
					}

					Index++;	/* increment the array index to search the next entry */

				} while(Result == TRUE);

	/* free the memory allocated for DetailData */
	if(detailData != NULL)
		free(detailData);

	/* free HID device info list resources */
	SetupDiDestroyDeviceInfoList(HidDevInfo);

	return FALSE;

}


/**********************************************************
*
*	Function: bHidDeviceNotify
*	Purpose: Sets up the HID device notification events.  The 
*			message contains the events
*			DBT_DEVICEARRIVAL and DBT_DEVICEREMOVALCOMPLETE.  It
*			is then up to the application to find out if the 
*			event is refering to the device it is connected with by
*			parsing a structure that the event points to.
*	Parameters: hWnd - handle to window that notifications should be
*				sent to.
*	Returns: TRUE, if device notification is set up
*			 FALSE, if device notification setup fails
*
**********************************************************/

BOOL bHidDeviceNotify(HWND hWnd, HDEVNOTIFY hDevNotify)
{
GUID HidGuid;										/* temporarily stores Windows HID Class GUID */
DEV_BROADCAST_DEVICEINTERFACE NotificationFilter;	/* un/plug notification filter */

	/* Set up device notification (i.e. plug or unplug of HID Devices) */
	
	/* 1) get the HID GUID */
	HidD_GetHidGuid(&HidGuid);
				
	/* 2) clear the notification filter */
	ZeroMemory( &NotificationFilter, sizeof(NotificationFilter));
				
	/* 3) assign the previously cleared structure with the correct data
	so that the application is notified of HID device un/plug events */
	NotificationFilter.dbcc_size = sizeof(DEV_BROADCAST_DEVICEINTERFACE);
	NotificationFilter.dbcc_devicetype = DBT_DEVTYP_DEVICEINTERFACE;
	NotificationFilter.dbcc_classguid = HidGuid;
				


	/* 4) register device notifications for this application */
	hDevNotify = RegisterDeviceNotification(	hWnd, 
												&NotificationFilter, 
												DEVICE_NOTIFY_WINDOW_HANDLE );
					
	/* 5) notify the calling procedure if the HID device will not be recognized */
	if(!hDevNotify)
		return FALSE;

	return TRUE;

}